/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * The contents of this file are subject to the Netscape Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/NPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is mozilla.org code.
 *
 * The Initial Developer of the Original Code is Netscape
 * Communications Corporation.  Portions created by Netscape are
 * Copyright (C) 1998 Netscape Communications Corporation. All
 * Rights Reserved.
 *
 * Contributor(s): 
 */

// wyciwyg implementation
// (borrows heavily from nsDiskCacheRecordChannel

#include "nsWyciwygChannel.h"
#include "nsIServiceManager.h"
#include "nsILoadGroup.h"
#include "nsIInterfaceRequestor.h"
#include "nsXPIDLString.h"
#include "nsISocketTransportService.h"
#include "nsIStringStream.h"
#include "nsMimeTypes.h"
#include "nsITXTToHTMLConv.h"
#include "netCore.h"
#include "nsIServiceManager.h"
#include "nsINetDataCacheManager.h"
#include "nsCacheManager.h"
#include "nsIScriptSecurityManager.h"
#include "nsNetUtil.h"
#include "nsIEventQueueService.h"

static NS_DEFINE_CID(kSocketTransportServiceCID, NS_SOCKETTRANSPORTSERVICE_CID);

#define BUFFER_SEG_SIZE (4*1024)
#define BUFFER_MAX_SIZE (64*1024)

// nsWyciwygChannel methods
nsWyciwygChannel::nsWyciwygChannel()
    : mLoadAttributes(nsIChannel::LOAD_NORMAL),
      mStatus(NS_OK)
{
    NS_INIT_REFCNT();
}

nsWyciwygChannel::~nsWyciwygChannel() {
}

NS_IMPL_THREADSAFE_ISUPPORTS4(nsWyciwygChannel, nsIChannel, nsIRequest,
                              nsIStreamListener, nsIStreamObserver)

nsresult
nsWyciwygChannel::Init(nsIURI* uri)
{
    nsresult rv;
    NS_ASSERTION(uri, "no uri");
    if (!uri)
        return NS_ERROR_NULL_POINTER;
    mURL = uri;

    nsXPIDLCString autoBuffer;
    rv = mURL->GetPath(getter_Copies(autoBuffer)); 
    if (NS_FAILED(rv)) return rv;

    // This autobuffer is the actual host for which this wyciwyg channel was 
    // created. We should set this to the mOriginalURI or vice versa
    return NS_OK;
}

NS_METHOD
nsWyciwygChannel::Create(nsISupports* aOuter, const nsIID& aIID, void* *aResult)
{
    nsWyciwygChannel* fc = new nsWyciwygChannel();
    if (fc == nsnull)
        return NS_ERROR_OUT_OF_MEMORY;
    NS_ADDREF(fc);
    nsresult rv = fc->QueryInterface(aIID, aResult);
    NS_RELEASE(fc);
    return rv;
}

///////////////////////////////////////////////////////////////////////////////
// nsIRequest methods:

NS_IMETHODIMP
nsWyciwygChannel::GetName(PRUnichar* *aName)
{
    NS_NOTREACHED("nsWyciwygChannel::GetName");
    return NS_ERROR_NOT_IMPLEMENTED;
}
 
NS_IMETHODIMP
nsWyciwygChannel::IsPending(PRBool *aIsPending)
{
  *aIsPending = PR_FALSE;
  if (!mCacheEntryChannel)
    return NS_OK;

  return mCacheEntryChannel->IsPending(aIsPending);
}

NS_IMETHODIMP
nsWyciwygChannel::GetStatus(nsresult *aStatus)
{
    *aStatus = mStatus;
    return NS_OK;
}

NS_IMETHODIMP
nsWyciwygChannel::Cancel(nsresult aStatus)
{
  NS_ASSERTION(NS_FAILED(aStatus), "shouldn't cancel with a success code");
  mStatus = aStatus;
  if (!mCacheEntryChannel)
    return NS_ERROR_FAILURE;

  return mCacheEntryChannel->Cancel(aStatus);
}
 
NS_IMETHODIMP
nsWyciwygChannel::Suspend(void)
{
  if (!mCacheEntryChannel)
    return NS_ERROR_FAILURE;

  return mCacheEntryChannel->Suspend();
}
 
NS_IMETHODIMP
nsWyciwygChannel::Resume(void)
{
  if (!mCacheEntryChannel)
    return NS_ERROR_FAILURE;

  return mCacheEntryChannel->Resume();
}

////////////////////////////////////////////////////////////////////////////////
// nsIChannel methods:

NS_IMETHODIMP
nsWyciwygChannel::GetOriginalURI(nsIURI* *aURI)
{
  // Let's hope this isn't called before mOriginalURI is set or we will
  // return the full wyciwyg URI for our originalURI  :S
  NS_ASSERTION(mOriginalURI, "nsWyciwygChannel::GetOriginalURI - mOriginalURI not set!\n");

  *aURI = mOriginalURI ? mOriginalURI : mURL;
  NS_ADDREF(*aURI);
  return NS_OK;
}

NS_IMETHODIMP
nsWyciwygChannel::SetOriginalURI(nsIURI* aURI)
{
  mOriginalURI = aURI;
  return NS_OK;
}

NS_IMETHODIMP
nsWyciwygChannel::GetURI(nsIURI* *aURI)
{
  // XXX Do we need to propagate this down to the cache entry channel?
  *aURI = mURL;
  NS_IF_ADDREF(*aURI);
  return NS_OK;
}

NS_IMETHODIMP
nsWyciwygChannel::SetURI(nsIURI* aURI)
{
    mURL = aURI;
    return NS_OK;
}

NS_IMETHODIMP
nsWyciwygChannel::OpenInputStream(nsIInputStream **aInputStream)
{
  nsresult rv = NS_OK;
  NS_ENSURE_ARG(aInputStream);

  NS_ASSERTION(!mCacheEntryChannel,
      "nsWyciwygChannel::OpenInputStream() called when wyciwyg channel was busy!\n");
  if (mCacheEntryChannel)
    return NS_ERROR_IN_PROGRESS;

  // Delegate to the cache entry channel
  GetCacheEntryChannel(PR_FALSE);

  if (mCacheEntryChannel)
    rv = mCacheEntryChannel->OpenInputStream(aInputStream);

  NS_ASSERTION(NS_SUCCEEDED(rv),
      "nsWyciwygChannel::OpenInputStream() failed!\n");
  if (NS_FAILED(rv))
    mCacheEntryChannel = null_nsCOMPtr();
  
  return rv;
}

NS_IMETHODIMP
nsWyciwygChannel::OpenOutputStream(nsIOutputStream **aOutputStream)
{
  nsresult rv = NS_OK;
  NS_ENSURE_ARG(aOutputStream);

  // This should basically set up a cache entry and get the output.
  // stream from it.  Writing to that stream goes straight to cache.

  NS_ASSERTION(!mCacheEntryChannel,
      "nsWyciwygChannel::OpenOutputStream() called when wyciwyg channel was busy!\n");
  if (mCacheEntryChannel)
    return NS_ERROR_IN_PROGRESS;

  // Delegate to the cache entry channel
  GetCacheEntryChannel(PR_TRUE);

  if (mCacheEntryChannel)
    rv = mCacheEntryChannel->OpenOutputStream(aOutputStream);

  NS_ASSERTION(NS_SUCCEEDED(rv),
      "nsWyciwygChannel::OpenOutputStream() failed!\n");
  if (NS_FAILED(rv))
    mCacheEntryChannel = null_nsCOMPtr();
  
  return rv;
}

NS_IMETHODIMP
nsWyciwygChannel::AsyncRead(nsIStreamListener *aListener, nsISupports *aContext)
{
  nsresult rv = NS_OK;

  if (mCacheEntryChannel)
    return NS_ERROR_IN_PROGRESS;


  // Add ourselves to the loadgroup so we get to proxy for the cache entry channel
  // We need to do this so that we can return a content type of text/html
  mRealListener = aListener;
  nsCOMPtr<nsIStreamListener> selfAsListener = this;

  if (mLoadGroup) {
    nsCOMPtr<nsILoadGroupListenerFactory> factory;
    //
    // Create a load group "proxy" listener...
    //
    rv = mLoadGroup->GetGroupListenerFactory(getter_AddRefs(factory));
    if (factory) {
      nsCOMPtr<nsIStreamListener> newListener;
      rv = factory->CreateLoadGroupListener(mRealListener, getter_AddRefs(newListener));
      if (NS_SUCCEEDED(rv)) {
        mRealListener = newListener;
      }
      // XXX WTF is the above failing?  Attempted hack below:
      // XXX Nowhere do I see a SetGroupListenerFactory...
    } else {
      nsCOMPtr<nsIStreamListener> newListener;
      NS_NewAsyncStreamListener(getter_AddRefs(newListener), mRealListener, NS_UI_THREAD_EVENTQ);
      if (NS_SUCCEEDED(rv)) {
        mRealListener = newListener;
      }
    }

    rv = mLoadGroup->AddChannel(this, nsnull);
    if (NS_FAILED(rv)) return rv;
  }

  // Delegate to the cache entry channel
  GetCacheEntryChannel(PR_FALSE);
  if (mCacheEntryChannel) {
    rv = mCacheEntryChannel->AsyncRead(selfAsListener, aContext);
    if (NS_FAILED(rv)) {
      // ResponseCompleted

      // First:
      //
      // Call the consumer OnStopRequest(...) to end the request...
      if (aListener)
      {
        rv = mRealListener->OnStopRequest(this, aContext, NS_OK, nsnull);
      }

      // Release the cache entry as soon as we are done. This helps as it can
      // flush any cache records and do maintenance.
      mCacheEntryChannel = null_nsCOMPtr();

      // After the consumer has been notified, remove the channel from its 
      // load group...  This will trigger an OnStopRequest from the load group.
      if (mLoadGroup)
        mLoadGroup->RemoveChannel(this, nsnull, NS_OK, nsnull);
    }
  }  

  return rv;
}

NS_IMETHODIMP
nsWyciwygChannel::AsyncWrite(nsIInputStream *aInputStream,
                            nsIStreamObserver *aObserver,
                            nsISupports *aContext)
{
    // Someday, maybe, like when nsCacheEntryChannel implements it.  :)
    NS_NOTREACHED("nsWyciwygChannel::AsyncWrite");
    return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP
nsWyciwygChannel::GetLoadAttributes(PRUint32 *aLoadAttributes)
{
    *aLoadAttributes = mLoadAttributes;
    return NS_OK;
}

NS_IMETHODIMP
nsWyciwygChannel::SetLoadAttributes(PRUint32 aLoadAttributes)
{
    mLoadAttributes = aLoadAttributes;
    return NS_OK;
}

// XXX Should we use "application/x-unknown-content-type" instead?  See nsDiskCacheRecordChannel.cpp
#define wyciwyg_TYPE TEXT_HTML

NS_IMETHODIMP
nsWyciwygChannel::GetContentType(char* *aContentType) {
    if (!aContentType) return NS_ERROR_NULL_POINTER;

    *aContentType = nsCRT::strdup(wyciwyg_TYPE);
    if (!*aContentType) return NS_ERROR_OUT_OF_MEMORY;
    return NS_OK;
}

NS_IMETHODIMP
nsWyciwygChannel::SetContentType(const char *aContentType)
{
  NS_NOTREACHED("nsWyciwygChannel::SetContentType");
  return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP
nsWyciwygChannel::GetContentLength(PRInt32 *aContentLength)
{
  NS_ASSERTION(mCacheEntryChannel,
      "nsWyciwygChannel::GetContentLength called before we have a cache channel!");
  if (mCacheEntryChannel) {
    return mCacheEntryChannel->GetContentLength(aContentLength);
  } else {
    return NS_ERROR_FAILURE;
  }
}

NS_IMETHODIMP
nsWyciwygChannel::SetContentLength(PRInt32 aContentLength)
{
  NS_NOTREACHED("nsWyciwygChannel::SetContentLength");
  return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP
nsWyciwygChannel::GetTransferOffset(PRUint32 *aTransferOffset)
{
  NS_ASSERTION(mCacheEntryChannel,
      "nsWyciwygChannel::GetTransferOffset called before we have a cache channel!");
  if (mCacheEntryChannel) {
    return mCacheEntryChannel->GetTransferOffset(aTransferOffset);
  } else {
    return NS_ERROR_FAILURE;
  }
}

NS_IMETHODIMP
nsWyciwygChannel::SetTransferOffset(PRUint32 aTransferOffset)
{
  NS_ASSERTION(mCacheEntryChannel,
      "nsWyciwygChannel::SetTransferOffset called before we have a cache channel!");
  if (mCacheEntryChannel) {
    return mCacheEntryChannel->SetTransferOffset(aTransferOffset);
  } else {
    return NS_ERROR_FAILURE;
  }
}

NS_IMETHODIMP
nsWyciwygChannel::GetTransferCount(PRInt32 *aTransferCount)
{
  NS_ASSERTION(mCacheEntryChannel,
      "nsWyciwygChannel::GetTransferCount called before we have a cache channel!");
  if (mCacheEntryChannel) {
    return mCacheEntryChannel->GetTransferCount(aTransferCount);
  } else {
    return NS_ERROR_FAILURE;
  }
}

NS_IMETHODIMP
nsWyciwygChannel::SetTransferCount(PRInt32 aTransferCount)
{
  NS_ASSERTION(mCacheEntryChannel,
      "nsWyciwygChannel::SetTransferCount called before we have a cache channel!");
  if (mCacheEntryChannel) {
    return mCacheEntryChannel->SetTransferCount(aTransferCount);
  } else {
    return NS_ERROR_FAILURE;
  }
}

NS_IMETHODIMP
nsWyciwygChannel::GetBufferSegmentSize(PRUint32 *aBufferSegmentSize)
{
  NS_ASSERTION(mCacheEntryChannel,
      "nsWyciwygChannel::GetBufferSegmentSize called before we have a cache channel!");
  if (mCacheEntryChannel) {
    return mCacheEntryChannel->GetBufferSegmentSize(aBufferSegmentSize);
  } else {
    return NS_ERROR_FAILURE;
  }
}

NS_IMETHODIMP
nsWyciwygChannel::SetBufferSegmentSize(PRUint32 aBufferSegmentSize)
{
  NS_ASSERTION(mCacheEntryChannel,
      "nsWyciwygChannel::SetBufferSegmentSize called before we have a cache channel!");
  if (mCacheEntryChannel) {
    return mCacheEntryChannel->SetBufferSegmentSize(aBufferSegmentSize);
  } else {
    return NS_ERROR_FAILURE;
  }
}

NS_IMETHODIMP
nsWyciwygChannel::GetBufferMaxSize(PRUint32 *aBufferMaxSize)
{
  NS_ASSERTION(mCacheEntryChannel,
      "nsWyciwygChannel::GetBufferMaxSize called before we have a cache channel!");
  if (mCacheEntryChannel) {
    return mCacheEntryChannel->GetBufferMaxSize(aBufferMaxSize);
  } else {
    return NS_ERROR_FAILURE;
  }
}

NS_IMETHODIMP
nsWyciwygChannel::SetBufferMaxSize(PRUint32 aBufferMaxSize)
{
  NS_ASSERTION(mCacheEntryChannel,
      "nsWyciwygChannel::SetBufferMaxSize called before we have a cache channel!");
  if (mCacheEntryChannel) {
    return mCacheEntryChannel->SetBufferMaxSize(aBufferMaxSize);
  } else {
    return NS_ERROR_FAILURE;
  }
}

NS_IMETHODIMP
nsWyciwygChannel::GetLocalFile(nsIFile* *aFile)
{
  *aFile = nsnull;
  return NS_OK;
}

NS_IMETHODIMP
nsWyciwygChannel::GetPipeliningAllowed(PRBool *aPipeliningAllowed)
{
    *aPipeliningAllowed = PR_FALSE;
    return NS_OK;
}
 
NS_IMETHODIMP
nsWyciwygChannel::SetPipeliningAllowed(PRBool aPipeliningAllowed)
{
    NS_NOTREACHED("SetPipeliningAllowed");
    return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP
nsWyciwygChannel::GetLoadGroup(nsILoadGroup* *aLoadGroup)
{
  *aLoadGroup = mLoadGroup;
  NS_IF_ADDREF(*aLoadGroup);
  return NS_OK;
}

NS_IMETHODIMP
nsWyciwygChannel::SetLoadGroup(nsILoadGroup* aLoadGroup)
{
  mLoadGroup = aLoadGroup;
  return NS_OK;
}

NS_IMETHODIMP
nsWyciwygChannel::GetOwner(nsISupports* *aOwner)
{
  nsresult rv = NS_OK;
  if (!mOwner) {
    // Create codebase principal with URI of original document, not our URI
    NS_ASSERTION(mOriginalURI,
        "nsWyciwygChannel::GetOwner without an owner or an original URI!");
    if (mOriginalURI) {
      nsIPrincipal* pIPrincipal = nsnull;
      NS_WITH_SERVICE(nsIScriptSecurityManager, secMan,
                     NS_SCRIPTSECURITYMANAGER_PROGID, &rv);
      if (secMan) {
        rv = secMan->GetCodebasePrincipal(mOriginalURI, &pIPrincipal);
        if (NS_SUCCEEDED(rv)) {
          mOwner = pIPrincipal;
          NS_RELEASE(pIPrincipal);
        }
      }
    } else {
      // Uh oh, must set originalURI before we can return an owner!
      return NS_ERROR_FAILURE;
    }
  }
  NS_ASSERTION(mOriginalURI,
      "nsWyciwygChannel::GetOwner unable to get owner!");
  if (mOwner) {
    *aOwner = mOwner.get();
    NS_IF_ADDREF(*aOwner);
  } else {
    *aOwner = nsnull;
  }
  return rv;
}

NS_IMETHODIMP
nsWyciwygChannel::SetOwner(nsISupports* aOwner)
{
  mOwner = aOwner;
  return NS_OK;
}

NS_IMETHODIMP
nsWyciwygChannel::GetNotificationCallbacks(nsIInterfaceRequestor* *aCallbacks)
{
  *aCallbacks = mCallbacks.get();
  NS_IF_ADDREF(*aCallbacks);
  return NS_OK;
}

NS_IMETHODIMP
nsWyciwygChannel::SetNotificationCallbacks(nsIInterfaceRequestor* aNotificationCallbacks)
{
  mCallbacks = aNotificationCallbacks;
  return NS_OK;
}

NS_IMETHODIMP 
nsWyciwygChannel::GetSecurityInfo(nsISupports * *aSecurityInfo)
{
  NS_ASSERTION(mCacheEntryChannel,
      "nsWyciwygChannel::GetSecurityInfo called before we have a cache channel!");
  if (mCacheEntryChannel) {
    return mCacheEntryChannel->GetSecurityInfo(aSecurityInfo);
  } else {
    return NS_ERROR_FAILURE;
  }
}

////////////////////////////////////////////////////////////////////////////////
// nsIStreamListener methods:
////////////////////////////////////////////////////////////////////////////////

NS_IMETHODIMP
nsWyciwygChannel::OnStartRequest(nsIChannel* aTransportChannel, nsISupports* aContext)
{
  NS_ASSERTION(mRealListener, "No listener...");
  return mRealListener->OnStartRequest(this, aContext);
}

NS_IMETHODIMP
nsWyciwygChannel::OnStopRequest(nsIChannel* TransportChannel, nsISupports* aContext,
                                nsresult aStatus, const PRUnichar* aStatusArg)
{
  nsresult rv;

  rv = mRealListener->OnStopRequest(this, aContext, aStatus, aStatusArg);

  if (mLoadGroup) {
    if (NS_SUCCEEDED(rv)) {
      mLoadGroup->RemoveChannel(this, aContext, aStatus, aStatusArg);
    }
  }

  // Release the reference to the consumer stream listener...
  mRealListener = null_nsCOMPtr();
  mCacheEntryChannel = null_nsCOMPtr();
  return rv;
}

NS_IMETHODIMP
nsWyciwygChannel::OnDataAvailable(nsIChannel* aTransportChannel, nsISupports* aContext,
                               nsIInputStream *aIStream, PRUint32 aSourceOffset,
                               PRUint32 aLength)
{
  nsresult rv;

  rv = mRealListener->OnDataAvailable(this, aContext, aIStream,
                                      aSourceOffset, aLength);

  // If the connection is being aborted cancel the transport.  This will
  // ensure that the transport will go away even if it is blocked waiting
  // for the consumer to empty the pipe...

  if (NS_FAILED(rv) && mCacheEntryChannel) {
    mCacheEntryChannel->Cancel(rv);
  }
  return rv;
}



////////////////////////////////////////////////////////////////////////////////
// Private helper method

nsresult
nsWyciwygChannel::GetCacheEntryChannel(PRBool aUpdateInProgress)
{
  nsresult rv = NS_OK;

  // check to see is the channel is just been used to write out to the 
  // cache. If that is the case then we shouldn't have to query the cache
  // if not then... query the cache!

  if (!mCacheEntryChannel) {

    // Get the cache manager service
    NS_WITH_SERVICE(nsINetDataCacheManager, cacheManager,
                    NS_NETWORK_CACHE_MANAGER_PROGID, &rv);
    NS_ASSERTION(NS_SUCCEEDED(rv), 
        "nsWyciwygChannel::GetCacheEntryChannel() You tried to use wyciwyg without a cache!\n");
    if (rv == NS_ERROR_FACTORY_NOT_REGISTERED) {
      return NS_OK; 
    } else if (NS_FAILED(rv))
      return rv;

    // ??? Force memory cache.
    PRUint32 cacheFlags = nsIChannel::INHIBIT_PERSISTENT_CACHING;

    // Retrieve an existing cache entry or create a new one if none 
    // exists for the given URL. (mURL is the wyciwyg URL)
    nsCOMPtr<nsICachedNetData> cacheEntry;
    nsXPIDLCString urlCString; 
    mURL->GetSpec(getter_Copies(urlCString));
    rv = cacheManager->GetCachedNetData(urlCString, 0, 0, cacheFlags,
                                        getter_AddRefs(cacheEntry));

    NS_ASSERTION(cacheEntry && NS_SUCCEEDED(rv),
                 "Cache manager must always return cache entry");

    // Due to architectural limitations in the cache manager, a cache entry can
    // not be written to if it is currently being read or written.  In that
    // case, we silently cancel the update of the cache entry.
    PRBool inUse;
    rv = cacheEntry->GetInUse(&inUse);
    if (NS_FAILED(rv)) return rv;
    if (inUse) {
#if defined(DEBUG_dp) || defined(DEBUG_neeti) || defined (DEBUG_gagan) || defined (DEBUG_pollmann)
        NS_ASSERTION(!inUse, "cacheEntry inUse. Cannot update with new content.");
#endif
        return NS_OK;
    }

    cacheEntry->SetUpdateInProgress(aUpdateInProgress);

    if (cacheEntry)
      rv = cacheEntry->NewChannel(mLoadGroup, getter_AddRefs(mCacheEntryChannel));
  }
  
  NS_ASSERTION(mCacheEntryChannel,
               "nsWyciwygChannel::GetCacheEntryChannel() failed!");
  if (mCacheEntryChannel) {
    // Propagate the load attributes of this channel into the cache channel.
    // This will ensure that notifications are suppressed if necessary.
    // XXX does this belong here? only in asyncRead?
    rv = mCacheEntryChannel->SetLoadAttributes(mLoadAttributes);


  }

  return rv;
}
